@@ -34,6 +34,16 @@ at any given time there's a valid logger available.
 
 Your MAIN module simply imports gmLog2 and all other modules
 will merrily and automagically start logging away.
+
+
+Ad hoc call stack logging recipe:
+
+	call_stack = inspect.stack()
+	call_stack.reverse()
+	for idx in range(1, len(call_stack)):
+		caller = call_stack[idx]
+		_log.debug('%s[%s] @ [%s] in [%s]', ' '* idx, caller[3], caller[2], caller[1])
+	del call_stack
 """
 # TODO:
 # - exception()
@@ -59,7 +69,7 @@ import calendar
 _logfile_name = None
 _logfile = None
 
-_string_encoding = None
+#_string_encoding = None
 
 # table used for cooking non-printables
 AsciiName = ['<#0-0x00-nul>',
@@ -121,32 +131,21 @@ AsciiName = ['<#0-0x00-nul>',
 #===============================================================
 def flush():
 	logger = logging.getLogger('gm.logging')
-	logger.critical(u'-------- synced log file -------------------------------')
+	logger.critical('-------- synced log file -------------------------------')
 	root_logger = logging.getLogger()
 	for handler in root_logger.handlers:
 		handler.flush()
 
-#===============================================================
-#def log_exception_details():
-#
-#	t, v, tb = sys.exc_info()
-#	if t is None:
-#		logger.debug('no exception to detail')
-#		return
-#
-#	logger = logging.getLogger('gm.logging')
-#	logger.debug('exception: %s', v)
-#	logger.debug('type: %s', t)
-#	logger.debug('list of attributes:')
-#	for attr in [ a for a in dir(v) if not a.startswith('__') ]:
-#		logger.debug('  %s: %s', attr, getattr(v, attr))
-
 #===============================================================
 def log_instance_state(instance):
 	logger = logging.getLogger('gm.logging')
 	logger.debug('state of %s', instance)
 	for attr in [ a for a in dir(instance) if not a.startswith('__') ]:
-		logger.debug('  %s: %s', attr, getattr(instance, attr))
+		try:
+			val = getattr(instance, attr)
+		except AttributeError:
+			val = '<cannot access>'
+		logger.debug('  %s: %s', attr, val)
 
 #===============================================================
 def log_stack_trace(message=None, t=None, v=None, tb=None):
@@ -160,21 +159,25 @@ def log_stack_trace(message=None, t=None, v=None, tb=None):
 	if tb is None:
 		tb = sys.exc_info()[2]
 	if tb is None:
-		logger.debug(u'sys.exc_info() did not return a traceback object, trying sys.last_traceback')
+		logger.debug('sys.exc_info() did not return a traceback object, trying sys.last_traceback')
 		try:
 			tb = sys.last_traceback
 		except AttributeError:
-			logger.debug(u'no stack to trace (no exception information available)')
+			logger.debug('no stack to trace (no exception information available)')
 			return
 
 	# log exception details
-	logger.debug(u'exception: %s', v)
-	logger.debug(u'type: %s', t)
-	logger.debug(u'list of attributes:')
-	for attr in [ a for a in dir(v) if not a.startswith(u'__') ]:
-		logger.debug(u'  %s: %s', attr, getattr(v, attr))
+	logger.debug('exception: %s', v)
+	logger.debug('type: %s', t)
+	logger.debug('list of attributes:')
+	for attr in [ a for a in dir(v) if not a.startswith('__') ]:
+		try:
+			val = getattr(v, attr)
+		except AttributeError:
+			val = '<cannot access>'
+		logger.debug('  %s: %s', attr, val)
 
-	# make sure we don't leave behind a bind
+	# make sure we don't leave behind a binding
 	# to the traceback as warned against in
 	# sys.exc_info() documentation
 	try:
@@ -195,59 +198,34 @@ def log_stack_trace(message=None, t=None, v=None, tb=None):
 
 	if message is not None:
 		logger.debug(message)
-	logger.debug(u'stack trace follows:')
-	logger.debug(u'(locals by frame, outmost frame first)')
+	logger.debug('stack trace follows:')
+	logger.debug('(locals by frame, outmost frame first)')
 	for frame in stack_of_frames:
 		logger.debug (
-			u'--- frame [%s]: #%s, %s -------------------',
+			'--- frame [%s]: #%s, %s -------------------',
 			frame.f_code.co_name,
 			frame.f_lineno,
 			frame.f_code.co_filename
 		)
 		for varname, value in frame.f_locals.items():
-			if varname == u'__doc__':
+			if varname == '__doc__':
 				continue
+			logger.debug('%20s = %s', varname, value)
 
-			try:
-				value = unicode(value, encoding = _string_encoding, errors = 'replace')
-			except TypeError:
-				try:
-					value = unicode(value)
-				except (UnicodeDecodeError, TypeError):
-					value = '%s' % str(value)
-					value = value.decode(_string_encoding, 'replace')
-
-			logger.debug(u'%20s = %s', varname, value)
-
-#===============================================================
-def set_string_encoding(encoding=None):
-
+#---------------------------------------------------------------
+def log_multiline(level, message=None, line_prefix=None, text=None):
+	if text is None:
+		return
+	if message is None:
+		message = 'multiline text:'
+	if line_prefix is None:
+		line_template = '  > %s'
+	else:
+		line_template = '%s: %%s' % line_prefix
+	lines2log = [message]
+	lines2log.extend([ line_template % line for line in text.split('\n') ])
 	logger = logging.getLogger('gm.logging')
-
-	global _string_encoding
-
-	if encoding is not None:
-		codecs.lookup(encoding)
-		_string_encoding = encoding
-		logger.info(u'setting python.str -> python.unicode encoding to <%s> (explicit)', _string_encoding)
-		return True
-
-	enc = sys.getdefaultencoding()
-	if enc != 'ascii':
-		_string_encoding = enc
-		logger.info(u'setting python.str -> python.unicode encoding to <%s> (sys.getdefaultencoding)', _string_encoding)
-		return True
-
-	enc = locale.getlocale()[1]
-	if enc is not None:
-		_string_encoding = enc
-		logger.info(u'setting python.str -> python.unicode encoding to <%s> (locale.getlocale)', _string_encoding)
-		return True
-
-	# FIXME: or rather use utf8 ?
-	_string_encoding = locale.getpreferredencoding(do_setlocale=False)
-	logger.info(u'setting python.str -> python.unicode encoding to <%s> (locale.getpreferredencoding)', _string_encoding)
-	return True
+	logger.log(level, '\n'.join(lines2log))
 
 #===============================================================
 # internal API
@@ -255,32 +233,33 @@ def set_string_encoding(encoding=None):
 __words2hide = []
 
 def add_word2hide(word):
+	if word is None:
+		return
+	if word.strip() == '':
+		return
 	if word not in __words2hide:
-		__words2hide.append(word)
+		__words2hide.append(str(word))
 
 #---------------------------------------------------------------
 __original_logger_write_func = None
 
 def __safe_logger_write_func(s):
 	for word in __words2hide:
-		# random is seeded from system time at import,
-		# jump ahead, scrambled by whatever is "now"
-		random.jumpahead(calendar.timegm(time.gmtime()))
-		# from that generate a replacement string valid for
+		# throw away up to 4 bits (plus the randint() cost)
+		random.getrandbits(random.randint(1, 4))
+		# from there generate a replacement string valid for
 		# *this* round of replacements of *this* word,
 		# this approach won't mitigate guessing trivial passwords
 		# from replacements of known data (a known-plaintext attack)
 		# but will make automated searching for replaced strings
 		# in the log more difficult
-		bummer = hex(random.randint(0, sys.maxint)).lstrip(u'0x')
+		bummer = hex(random.randint(0, sys.maxsize)).lstrip('0x')
 		s = s.replace(word, bummer)
 	__original_logger_write_func(s)
 
 #---------------------------------------------------------------
 def __setup_logging():
 
-	set_string_encoding()
-
 	global _logfile
 	if _logfile is not None:
 		return True
@@ -288,33 +267,31 @@ def __setup_logging():
 	if not __get_logfile_name():
 		return False
 
-	if sys.version[:3] < '2.5':
-		fmt = u'%(asctime)s  %(levelname)-8s  %(name)-12s  (%(pathname)s @ #%(lineno)d): %(message)s'
-	else:
-		fmt = u'%(asctime)s  %(levelname)-8s  %(name)-12s  [%(thread)d %(threadName)-10s]  (%(pathname)s::%(funcName)s() #%(lineno)d): %(message)s'
-
 	_logfile = io.open(_logfile_name, mode = 'wt', encoding = 'utf8', errors = 'replace')
-
 	global __original_logger_write_func
 	__original_logger_write_func = _logfile.write
 	_logfile.write = __safe_logger_write_func
 
+	# setup
+	fmt = '%(asctime)s  %(levelname)-8s  %(name)-12s  [%(thread)d %(threadName)-10s]  (%(pathname)s::%(funcName)s() #%(lineno)d): %(message)s'
 	logging.basicConfig (
 		format = fmt,
 		datefmt = '%Y-%m-%d %H:%M:%S',
 		level = logging.DEBUG,
 		stream = _logfile
 	)
-
 	logging.captureWarnings(True)
+	logger = logging.getLogger()
+	logger.log_stack_trace = log_stack_trace
+	logger.log_multiline = log_multiline
 
-	logger = logging.getLogger('gm.logging')
-	logger.critical(u'-------- start of logging ------------------------------')
-	logger.info(u'log file is <%s>', _logfile_name)
-	logger.info(u'log level is [%s]', logging.getLevelName(logger.getEffectiveLevel()))
-	logger.info(u'log file encoding is <utf8>')
-	logger.info(u'initial python.str -> python.unicode encoding is <%s>', _string_encoding)
-	logger.debug(u'log file .write() patched from original %s to patched %s', __original_logger_write_func, __safe_logger_write_func)
+	# start logging
+	#logger = logging.getLogger('gm.logging')
+	logger.critical('-------- start of logging ------------------------------')
+	logger.info('log file is <%s>', _logfile_name)
+	logger.info('log level is [%s]', logging.getLevelName(logger.getEffectiveLevel()))
+	logger.info('log file encoding is <utf8>')
+	logger.debug('log file .write() patched from original %s to patched %s', __original_logger_write_func, __safe_logger_write_func)
 
 #---------------------------------------------------------------
 def __get_logfile_name():
@@ -332,11 +309,11 @@ def __get_logfile_name():
 
 	# given on command line ?
 	for option in sys.argv[1:]:
-		if option.startswith(u'--log-file='):
-			(opt_name, value) = option.split(u'=')
+		if option.startswith('--log-file='):
+			(opt_name, value) = option.split('=')
 			(dir_name, file_name) = os.path.split(value)
-			if dir_name == u'':
-				dir_name = u'.'
+			if dir_name == '':
+				dir_name = '.'
 			if file_name == '':
 				file_name = default_logfile_name
 			_logfile_name = os.path.abspath(os.path.expanduser(os.path.join(dir_name, file_name)))
@@ -364,20 +341,23 @@ if __name__ == '__main__':
 	if len(sys.argv) < 2:
 		sys.exit()
 
-	if sys.argv[1] != u'test':
+	if sys.argv[1] != 'test':
 		sys.exit()
 
 	#-----------------------------------------------------------
 	def test():
 		logger = logging.getLogger('gmLog2.test')
+
+		logger.error('test %s', [1,2,3])
+
 		logger.error("I expected to see %s::test()" % __file__)
-		add_word2hide(u'super secret passphrase')
+		add_word2hide('super secret passphrase')
 		logger.debug('credentials: super secret passphrase')
 
 		try:
 			int(None)
-		except:
-			logger.exception(u'unhandled exception')
+		except Exception:
+			logger.exception('unhandled exception')
 			log_stack_trace()
 		flush()
 	#-----------------------------------------------------------
