@@ -1,7 +1,6 @@
 -- find duplicate streets
 select
-	id,
-	name,
+	id, name,
 	postcode
 from
 	dem.street
